"""Requests related data models."""
from dataclasses import dataclass
from dataclasses import field
from datetime import datetime
from enum import Enum

from bosch.control_panel.cc880p.utils import checksum
from bosch.control_panel.cc880p.utils import swap_nibbles


class RequestsProps(Enum):
    """Requests related properties."""

    StatusRequestCode = '01'
    StatusRequestSize = 11
    KeysRequestCode = '0C'
    KeysRequestSize = 11
    SetOutRequestCode = '0E'
    SetOutRequestSize = 11
    SetArmingRequestCode = '0E'
    SetArmingRequestSize = 11
    SetSirenRequestCode = '0E'
    SetSirenRequestSize = 11
    SetTimeRequestCode = '0E'
    SetTimeRequestSize = 11


@dataclass
class Request:
    """Base Request Class."""

    code: str
    size: int

    def encode(self, *args, **kwargs):
        """Encode the request. Should be overridden by child classes."""
        raise NotImplementedError('Implement the encode method')

    def _encode(self, data: bytes) -> bytes:
        """Encode bytes adding the request code and the checksum."""
        _data = bytes.fromhex(self.code) + data
        return _data + bytes([checksum(_data)])


@dataclass
class KeysRequest(Request):
    """Keys Request."""

    code: str = field(init=False, default=RequestsProps.KeysRequestCode.value)
    size: int = field(init=False, default=RequestsProps.KeysRequestSize.value)
    keys: bytes
    zone: int

    def encode(self):
        """Encode the Keys sending request."""
        # Remaining bytes after the keys should be 0
        empty = bytes([0x00] * (7 - len(self.keys)))
        zone = bytes([self.zone])
        n_keys = bytes([len(self.keys)])
        return self._encode(self.keys + empty + zone + n_keys)


@dataclass
class StatusRequest(Request):
    """Status Request."""

    code: str = field(
        init=False,
        default=RequestsProps.StatusRequestCode.value)
    size: int = field(
        init=False,
        default=RequestsProps.StatusRequestSize.value)
    installer_code: str

    def encode(self):
        """Encode the status request."""
        empty3 = bytes([0x00] * 3)
        empty2 = bytes([0x00] * 2)

        return self._encode(empty3 + self._encode_code() + empty2)

    def _encode_code(self):
        _code = self.installer_code + \
            ('f' * (7 - len(self.installer_code))) + '0'
        _code_bytes = bytes.fromhex(_code)
        return bytes(swap_nibbles(b) for b in _code_bytes)


@dataclass
class SetOutRequest(Request):
    """Set Output Request."""

    code: str = field(
        init=False,
        default=RequestsProps.SetOutRequestCode.value)
    size: int = field(
        init=False,
        default=RequestsProps.SetOutRequestSize.value)
    out_id: int
    on: bool

    def encode(self):
        """Encode the set output request."""
        empty7 = [0x00] * 7
        status = 0x03 if self.on else 0x04
        idx = self.out_id - 1
        return self._encode(bytes([status, idx, *empty7]))


@dataclass
class SetArmingRequest(Request):
    """Set Arming Request."""

    code: str = field(
        init=False,
        default=RequestsProps.SetArmingRequestCode.value)
    size: int = field(
        init=False,
        default=RequestsProps.SetArmingRequestSize.value)
    area_id: int
    arm: bool

    def encode(self):
        """Encode the set arming request."""
        empty7 = [0x00] * 7
        status = 0x01 if self.arm else 0x02
        idx = self.area_id - 1
        return self._encode(bytes([status, idx, *empty7]))


@dataclass
class SetSirenRequest(Request):
    """Set Siren Request."""

    code: str = field(
        init=False,
        default=RequestsProps.SetSirenRequestCode.value)
    size: int = field(
        init=False,
        default=RequestsProps.SetSirenRequestSize.value)
    on: bool

    def encode(self):
        """Encode the request to set the siren on/off."""
        empty8 = [0x00] * 8
        status = 0x05 if self.on else 0x06
        return self._encode(bytes([status, *empty8]))


@dataclass
class SetTimeRequest(Request):
    """Set TIme Request."""

    code: str = field(
        init=False,
        default=RequestsProps.SetSirenRequestCode.value)
    size: int = field(
        init=False,
        default=RequestsProps.SetSirenRequestSize.value)
    time: datetime

    def encode(self):
        """Encode the request to set the siren on/off."""
        empty3 = [0x00] * 3
        set_time = 0x0C

        if self.time is None:
            self.time = datetime.now()

        year = int(self.time.strftime('%y'))
        return self._encode(bytes([
            set_time,
            self.time.hour,
            self.time.minute,
            year,
            self.time.month,
            self.time.day,
            *empty3
        ]))



"""Control panel models."""
import datetime
from dataclasses import dataclass
from enum import Enum
from typing import Dict

from bosch.control_panel.cc880p.models.constants import Id


@dataclass
class CpModel:
    """Control Panel model properties."""

    n_zones: int
    n_areas: int
    n_outputs: int


class CpVersion(Enum):
    """Control Panel Version to be used.

    Returns a Control Panel model object with the properties referred to
    the selected version
    """

    S16_V14 = CpModel(
        n_zones=16,
        n_areas=1,
        n_outputs=5
    )

    def model(self) -> CpModel:
        """Return the CpModel object."""
        return self.value


@dataclass
class ControlPanelEntity:
    """Model representing a Control Panel entity."""


@dataclass
class Siren(ControlPanelEntity):
    """Dataclass to store the Siren of the alarm."""

    on: bool = False


@dataclass
class Zone(ControlPanelEntity):
    """Dataclass to store the zones of the alarm."""

    triggered: bool = False
    enabled: bool = False


@dataclass
class Output(ControlPanelEntity):
    """Dataclass to store the various alarm output states."""

    on: bool = False


@dataclass
class Availability(ControlPanelEntity):
    """Dataclass to store the various alarm output states."""

    available: bool = False


class ArmingMode(Enum):
    """Enumerator with all the alarm states."""

    DISARMED = 0
    ARMED_AWAY = 1
    ARMED_STAY = 2


@dataclass
class Area(ControlPanelEntity):
    """Dataclass representing the alarm area."""

    mode: ArmingMode = ArmingMode.DISARMED


@dataclass
class Time(ControlPanelEntity):
    """Datetime."""

    time: datetime.time = datetime.time(hour=0, minute=0)

    def __str__(self) -> str:
        """Convert the object hours and minutes into a string."""
        return f'{self.time.hour}:{self.time.minute}'

    def __repr__(self) -> str:
        """Convert the Time object into its string representation."""
        return f'Time(time={str(self)})'


@dataclass
class ControlPanel(ControlPanelEntity):
    """Dataclass representing the control panel object."""

    siren: Siren
    areas: Dict[Id, Area]
    zones: Dict[Id, Zone]
    outputs: Dict[Id, Output]
    time: Time
    availability: Availability

    @staticmethod
    def build(model: CpModel) -> 'ControlPanel':
        """Build a control panel object."""
        return ControlPanel(
            siren=Siren(),
            outputs={i + 1: Output() for i in range(model.n_outputs)},
            areas={i + 1: Area() for i in range(model.n_areas)},
            zones={i + 1: Zone() for i in range(model.n_zones)},
            time=Time(),
            availability=Availability()
        )



"""Command Line Commands Management."""
from datetime import datetime
from typing import Any
from typing import List

from bosch.control_panel.cc880p.cli.parser import Args
from bosch.control_panel.cc880p.cli.parser import Commands
from bosch.control_panel.cc880p.cp import CP
from bosch.control_panel.cc880p.utils import checksum


def cmd(cp: CP, obj: Any = None):
    """Fetch the control panel status after any change."""
    def inner(func):
        async def wrapper(*args, **kwargs):
            # await cp.get_status()
            if obj is not None:
                print(f'Before: {obj}')
            await func(*args, **kwargs)
            # await cp.get_status()
            if obj is not None:
                print(f'After: {obj}')
        return wrapper
    return inner


async def set_mode_arming(cp: CP, arm: bool):
    """Handle the arming mode command."""
    await cmd(cp, cp.control_panel.areas[1])(cp.set_arming)(arm=arm)


async def handle_mode_command(cp: CP, mode: str):
    """Handle the mode command."""
    if mode == 'arm':
        await set_mode_arming(cp, True)
    elif mode == 'disarm':
        await set_mode_arming(cp, False)


async def handle_siren_command(cp: CP, status: bool):
    """Handle the siren command."""
    await cmd(cp, cp.control_panel.siren)(cp.set_siren)(on=status)


async def handle_out_command(cp: CP, out: int, status: bool):
    """Handle outputs command."""
    await cmd(cp, cp.control_panel.outputs[out])(cp.set_output)(
        id=out,
        on=status
    )


async def handle_time_command(cp: CP, time: datetime):
    """Handle the set time command."""
    await cmd(cp, cp.control_panel.time)(cp.set_time)(time=time)


async def handle_keys_command(cp: CP, keys: List[str]):
    """Handle the keys sending."""
    keys = [i for ele in keys for i in ele]
    await cmd(cp)(cp.send_keys)(keys=keys)
    await cmd(cp, cp.control_panel.areas)(cp.get_status)()


async def handle_raw_command(cp: CP, raw: bytes):
    """Handle raw commands."""
    # Add the checksum calculated to the data
    _raw = raw + bytes([checksum(raw)])
    print(f'Sending: {_raw.hex()}')
    # Send the command and wait for the response
    resp = await (cp.send_command)(message=_raw)

    # Split the response checksum from its data
    resp_checksum = resp[-1]
    resp_data = resp[0:-1]
    # Calculate the checksum
    calc_checksum = checksum(resp_data)

    if calc_checksum == resp_checksum:
        print(
            f'[VALID CHECKSUM] received: {hex(resp_checksum)} | calculated: '
            f'{hex(calc_checksum)}'
        )
    else:
        print(
            f'[INVALID CHECKSUM] received: {hex(resp_checksum)} | calculated: '
            f'{hex(calc_checksum)}'
        )

    print(resp.hex())


async def handle_command(cp: CP, args: Args):
    """Handle the control panel commands."""
    if args.cmd == Commands.SetMode:
        await handle_mode_command(cp, args.mode)
    if args.cmd == Commands.SetSiren:
        await handle_siren_command(cp, args.status)
    if args.cmd == Commands.SetTime:
        await handle_time_command(cp, args.time)
    if args.cmd == Commands.SetOutput:
        await handle_out_command(cp, args.out, args.status)
    if args.cmd == Commands.SendKeys:
        await handle_keys_command(cp, args.keys)
    if args.cmd == Commands.SendRaw:
        await handle_raw_command(cp, args.raw)



"""Utilities."""
from typing import Tuple
from typing import Union


def odd_parity(n: int) -> bool:
    """Calculate the odd parity of a certain byte."""
    y = n ^ (n >> 1)
    y = y ^ (y >> 2)
    y = y ^ (y >> 4)
    y = y ^ (y >> 8)
    y = y ^ (y >> 16)

    if (y & 1):
        return True
    return False


def parities(data: bytes) -> Tuple[int, int]:
    """Calculate the number of odd and even parities."""
    odds = 0
    evens = 0
    for d in data:
        if odd_parity(d):
            odds += 1
        else:
            evens += 1

    return odds, evens


def evens(data: bytes) -> int:
    """Return the number of even parities."""
    _, evens = parities(data)
    return evens


def odds(data: bytes) -> int:
    """Return the number of odd parities."""
    odds, _ = parities(data)
    return odds


def checksum(data: bytes) -> int:
    """Calculate the checksum of a set of bytes."""
    return (sum(data) + evens(data)) & 0xFF


def swap_nibbles(x):
    """Swap byte nibbles."""
    return ((x & 0x0F) << 4 | (x & 0xF0) >> 4)


def to_hex(_bytes: Union[bytes, int]) -> str:
    """Convert a single byte or a list of bytes into hex format.

    Args:
        _bytes (bytes|int): Single byte or a list of bytes

    Returns:
        str: String representation of the byte(s) in hexadecimal format
    """
    _byte_str = ''
    if isinstance(_bytes, int):
        _bytes = bytes([_bytes])

    for _byte in _bytes:
        _byte_str += f'{int(_byte):02x}'
        _byte_str += ' '
    return _byte_str


def to_bin(_bytes: Union[bytes, int]):
    """Convert a single byte or a list of bytes into binary format.

    Args:
        _bytes (bytes|int): Single byte or a list of bytes

    Returns:
        str: String representation of the byte(s) in binary format
    """
    _byte_str = ''
    if isinstance(_bytes, int):
        _bytes = bytes([_bytes])

    for _byte in _bytes:
        _byte_str += f'{int(_byte):08b}'
        _byte_str += ' '
    return _byte_str



"""Responses data models."""
import datetime
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import List

from bosch.control_panel.cc880p.models.cp import ArmingMode
from bosch.control_panel.cc880p.models.cp import CpModel


class ResponsesProps(Enum):
    """Responses properties."""

    StatusCode = '04'
    StatusSize = 13


@dataclass
class Response:
    """Response base object."""

    code: str
    size: int


@dataclass
class StatusResponse:
    """Status response object."""

    code: str = field(init=False, default=ResponsesProps.StatusCode.value)
    size: int = field(init=False, default=ResponsesProps.StatusSize.value)
    time: datetime.time
    siren: bool
    outs: List[bool]
    areas: List[ArmingMode]
    zones: List[bool]
    zones_en: List[bool]

    @classmethod
    def decode(cls, data: bytes, cp: CpModel):
        """Decode a status response."""
        time = cls._decode_time(data)
        siren = cls._decode_siren_status(data)
        outs = cls._decode_outputs(data, cp.n_outputs)
        areas = cls._decode_areas(data, cp.n_areas)
        zones = cls._decode_zones(data, cp.n_zones)
        zones_en = cls._decode_zones_en(data, cp.n_zones, areas[0])

        return StatusResponse(
            time=time,
            siren=siren,
            outs=outs,
            areas=areas,
            zones=zones,
            zones_en=zones_en
        )

    @classmethod
    def _decode_time(cls, data: bytes) -> datetime.time:
        bytes = data[10:12]
        # Hours
        hours = bytes[0] & 0x1F  # Only the first 5 bits matters (0h-23h)
        # Minutes
        minutes = bytes[1] & 0x3F  # Only the first 6 bits matters (0m-59m)
        # Time
        return datetime.time(hour=hours, minute=minutes)

    @classmethod
    def _decode_siren_status(cls, data: bytes) -> bool:
        _byte = data[10]
        bit = 6

        return bool(_byte & (1 << bit))

    @classmethod
    def _decode_outputs(cls, data: bytes, n_outputs: int):
        outs: List[bool] = []
        bytes = data[1:3]

        for i in range(n_outputs):
            bit = i % 8
            byte = len(bytes) - int(i / 8) - 1

            outs.append(bool(bytes[byte] & (1 << bit)))

        return outs

    @classmethod
    def _decode_areas(cls, data: bytes, n_areas: int):
        areas: List[ArmingMode] = []
        _byte = data[9]

        for i in range(n_areas):
            away_bit = i % 4
            stay_bit = away_bit + 4
            away_status = bool(_byte & (1 << away_bit))
            stay_status = bool(_byte & (1 << stay_bit))

            if away_status:
                area = ArmingMode.ARMED_AWAY
            elif stay_status:
                area = ArmingMode.ARMED_STAY
            else:
                area = ArmingMode.DISARMED

            areas.append(area)

        return areas

    @classmethod
    def _decode_zones(cls, data: bytes, n_zones: int):
        bytes = data[3:5]
        zones: List[bool] = []

        for i in range(n_zones):
            bit = i % 8
            byte = int(i / 8)

            zones.append(bool(bytes[byte] & (1 << bit)))

        return zones

    @classmethod
    def _decode_zones_en(cls, data: bytes, n_zones: int, area: ArmingMode):
        bytes = data[5:7]
        zones_en: List[bool] = []
        status = False

        if area == ArmingMode.ARMED_AWAY:
            status = True
        elif area == ArmingMode.DISARMED:
            status = False

        for i in range(n_zones):
            bit = i % 8
            byte = int(i / 8)

            if area == ArmingMode.ARMED_STAY:
                status = bool(bytes[byte] & (1 << bit))

            zones_en.append(status)

        return zones_en



"""Control panel command line parser."""
from argparse import ArgumentParser
from datetime import datetime
from enum import Enum
from typing import List
from typing import Optional

from bosch.control_panel.cc880p.version import __version__ as version
from distutils.util import strtobool


class Commands(Enum):
    """Supported Commands."""

    SetSiren = 0
    SetMode = 1
    SetOutput = 2
    SetTime = 3
    SendKeys = 4
    SendRaw = 5


class Args:
    """Parser arguments."""

    connect: str
    port: int
    cmd: Optional[Commands]
    keys: List[str]
    mode: str
    status: bool
    out: int
    time: datetime


args = Args()


def get_parser():
    """Get the argument parser."""
    parser = ArgumentParser(description='Connects to the Control Panel')
    parser.add_argument(
        'cmd',
        action='store_const',
        const=None
    )
    parser.add_argument(
        '-v', '--version',
        action='version',
        version=version,
        help='Gets the current version'
    )
    parser.add_argument(
        '-c', '--connect',
        required=True,
        metavar='IP',
        type=str,
        help='the host ip'
    )
    parser.add_argument(
        '-p', '--port',
        required=True,
        metavar='PORT',
        type=int,
        help='the host port'
    )

    parser.add_argument(
        '-i', '--code',
        required=False,
        metavar='INSTALLER_CODE',
        type=str,
        help='the installer code'
    )

    subparsers = parser.add_subparsers()

    get_cmds_parser(subparsers)

    return parser


def get_cmds_parser(subparsers):
    """Get commands parser."""
    # Command Parser
    cmds_parser = subparsers.add_parser('cmd', help='Execute a command')

    subparsers = cmds_parser.add_subparsers()

    get_cmd_send_keys_parser(subparsers)
    get_cmd_send_raw_parser(subparsers)
    get_cmd_set_mode_parser(subparsers)
    get_cmd_set_siren_parser(subparsers)
    get_cmd_set_output_parser(subparsers)
    get_cmd_set_time_parser(subparsers)


def get_cmd_send_keys_parser(subparsers):
    """Send keys command parser."""
    # SEND KEYS command
    cmd_send_keys_parser = subparsers.add_parser(
        'sendKeys',
        help='Sends a set of keys to the control panel. Currently supports the'
        ' following: [0-9*#]{1,7}'
    )
    cmd_send_keys_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SendKeys
    )
    cmd_send_keys_parser.add_argument(
        'keys',
        nargs='+',
        type=str,
        help='The keys to execute'
    )


def get_cmd_send_raw_parser(subparsers):
    """Send raw command parser."""
    # SEND Raw command
    cmd_send_raw_parser = subparsers.add_parser(
        'sendRaw',
        help='Sends a set of bytes to the control panel.'
    )
    cmd_send_raw_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SendRaw
    )

    def string_to_bytes(string: str) -> bytes:
        return bytes.fromhex(string)

    cmd_send_raw_parser.add_argument(
        'raw',
        type=lambda string: bytes.fromhex(string),
        help='The byte to send'
    )


def get_cmd_set_mode_parser(subparsers):
    """Set mode command parser."""
    # SET MODE command
    cmd_set_mode_parser = subparsers.add_parser(
        'setMode',
        help='Change the control panel mode like arm, disarm, etc'
    )
    cmd_set_mode_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SetMode
    )
    cmd_set_mode_parser.add_argument(
        'mode',
        choices=['arm', 'disarm'],
        help='Changes the control panel mode'
    )


def get_cmd_set_siren_parser(subparsers):
    """Set siren command parser."""
    cmd_set_siren_parser = subparsers.add_parser(
        'setSiren',
        help='Change the control panel siren status'
    )
    cmd_set_siren_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SetSiren
    )
    cmd_set_siren_parser.add_argument(
        'status',
        type=lambda x: bool(strtobool(x)),
        help='Changes the status of the siren'
    )


def get_cmd_set_output_parser(subparsers):
    """Set output parser."""
    # SET OUTPUT command
    cmd_set_out_parser = subparsers.add_parser(
        'setOut',
        help='Change the output status of the control panel'
    )
    cmd_set_out_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SetOutput
    )
    cmd_set_out_parser.add_argument(
        'out',
        type=int,
        help='Changes the output'
    )
    cmd_set_out_parser.add_argument(
        'status',
        type=lambda x: bool(strtobool(x)),
        help='Changes the output'
    )


def get_cmd_set_time_parser(subparsers):
    """Set output parser."""
    # SET OUTPUT command
    cmd_set_time_parser = subparsers.add_parser(
        'setTime',
        help='Set the time of the control panel'
    )
    cmd_set_time_parser.add_argument(
        'cmd',
        action='store_const',
        const=Commands.SetTime
    )
    cmd_set_time_parser.add_argument(
        '-t', '--time',
        required=False,
        metavar='TIME',
        type=datetime.fromisoformat,
        help=(
            'Datetime to be set on the control panel. It should be in iso '
            'format (e.g.: "2023-12-31 23:59:59"). If not used, the current '
            'time will be applied.'
        )
    )


def get_args():
    """Parse and get the arguments."""
    return get_parser().parse_args(namespace=args)



"""Control Panel."""
import asyncio
import dataclasses
import datetime
import logging
from asyncio import AbstractEventLoop
from asyncio import Task
from asyncio.exceptions import CancelledError
from enum import Enum
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from bosch.control_panel.cc880p.models.callbacks import ControlPanelListener
from bosch.control_panel.cc880p.models.callbacks import DataListener
from bosch.control_panel.cc880p.models.constants import Id
from bosch.control_panel.cc880p.models.constants import MAX_KEYS
from bosch.control_panel.cc880p.models.cp import ArmingMode
from bosch.control_panel.cc880p.models.cp import ControlPanel
from bosch.control_panel.cc880p.models.cp import CpModel
from bosch.control_panel.cc880p.models.requests import KeysRequest
from bosch.control_panel.cc880p.models.requests import Request
from bosch.control_panel.cc880p.models.requests import SetArmingRequest
from bosch.control_panel.cc880p.models.requests import SetOutRequest
from bosch.control_panel.cc880p.models.requests import SetSirenRequest
from bosch.control_panel.cc880p.models.requests import SetTimeRequest
from bosch.control_panel.cc880p.models.requests import StatusRequest
from bosch.control_panel.cc880p.models.responses import Response
from bosch.control_panel.cc880p.models.responses import StatusResponse
from bosch.control_panel.cc880p.utils import checksum
from bosch.control_panel.cc880p.utils import to_hex

_LOGGER = logging.getLogger(__name__)


class CP:
    """Alarm control panel object representation."""

    def __init__(
        self,
        ip: str,
        port: int,
        model: CpModel,
        installer_code: str = None,
        loop: Optional[AbstractEventLoop] = None,
    ):
        """Initialize the Control Panel object to interface through TCP.

        Args:
            ip (str):
                IP of the control panel
            port (int):
                Port of the control panel
            mode (CpModel):
                Control Panel Model.
            loop (AbstractEventLoop, optional):
                Event Loop. Defaults None.
        """
        # Main event loop
        self._loop = loop or asyncio.get_event_loop()

        # Control Panel model
        self._model = model

        # Installer code needed to fetch control panel statuses
        self._installer_code = installer_code

        # IP of the control panel
        self._ip = ip
        # Port of the control panel
        self._port = port

        # Streamreader
        self._reader: Optional[asyncio.StreamReader] = None
        # Streamwriter
        self._writer: Optional[asyncio.StreamWriter] = None
        # Connected
        self._connected: bool = False
        # Reconnection task
        self._conn_task: Optional[Task] = None

        # Lock
        self._lock = asyncio.Lock()

        # Listeners to be called whenever the control panel state is changed
        self._control_panel_listeners: List[ControlPanelListener] = []
        # Listeners to be called whenever the there's new data
        self._data_listeners: List[DataListener] = []

        self._control_panel = ControlPanel.build(self._model)

    def __repr__(self):
        """Representation of control panel."""
        return str(self._control_panel)

    @property
    def __dict__(self):
        """Return a dict representation of control panel."""
        def _custom_asdict_factory(data):
            def convert_value(obj):
                if isinstance(obj, Enum):
                    return obj.name
                return obj

            return {k: convert_value(v) for k, v in data}

        return dataclasses.asdict(
            self._control_panel,
            dict_factory=_custom_asdict_factory
        )

    @property
    def control_panel(self) -> ControlPanel:
        """Property that returns the control panel object."""
        return self._control_panel

    @property
    def connected(self) -> bool:
        """Report whether the connection to the alarm is established."""
        return self._connected

    async def start(self) -> 'CP':
        """Establish the connection to the control panel."""
        await self._start_connection_task()
        return self

    async def stop(self) -> 'CP':
        """Stop the connection to the control panel."""
        await self._stop_connection_task()
        return self

    def add_control_panel_listener(self, listener: ControlPanelListener):
        """Add a listener function to listen for any change in the alarm."""
        self._control_panel_listeners.append(listener)

    def add_data_listener(self, listener: DataListener):
        """Add a listener function to listen for any incoming data."""
        self._data_listeners.append(listener)

    async def send_keys(
        self,
        keys: Union[str, List[str]],
    ):
        """Simulate a keypad, allowing sending multiple keys."""
        encoded_keys = self._encode_keys(keys)

        for i in range(0, len(encoded_keys), MAX_KEYS):
            # Get the next keys subset
            keys_subset = encoded_keys[i: i + MAX_KEYS]
            # Build the request
            request = KeysRequest(keys=keys_subset, zone=0)
            # Send the request
            await self.send_request(request, StatusResponse)

    async def set_output(self, id: Id, on: bool):
        """Set output."""
        try:
            if id not in self.control_panel.outputs:
                raise ValueError(f"The output with {id} doesn't exist")

            if self._control_panel.outputs[id].on != on:
                request = SetOutRequest(id, on)
                await self.send_request(request, StatusResponse)
        except Exception as ex:
            _LOGGER.error(f'Error setting the output: {ex}')
            raise

    async def set_arming(
        self,
        id: Id = 1,
        arm: bool = False
    ):
        """Set arming."""
        if arm and self._control_panel.areas[id].mode == ArmingMode.DISARMED:
            request = SetArmingRequest(id, arm)
            await self.send_request(request, StatusResponse)

        elif not arm and \
                self._control_panel.areas[id].mode != ArmingMode.DISARMED:
            request = SetArmingRequest(id, arm)
            await self.send_request(request, StatusResponse)

    async def set_siren(self, on: bool = False):
        """Set siren on or off.."""
        if on and self._control_panel.siren.on != on:
            # Switch on the siren
            await self.send_request(SetSirenRequest(on), StatusResponse)
        elif not on and self._control_panel.siren.on != on:
            # Switch of the siren
            await self.send_request(SetSirenRequest(on), StatusResponse)

    async def set_time(self, time: datetime.datetime = None):
        """Set time."""
        # Set the time in the alarm
        await self.send_request(SetTimeRequest(time), StatusResponse)

    async def get_status(self):
        """Command to request the status of the alarm."""
        request = StatusRequest(installer_code=self._installer_code)
        await self.send_request(request, StatusResponse)

    async def send_request(
            self,
            request: Request,
            resp_type: Optional[Type[Response]] = None
    ):
        """Send a new request."""
        n_resp_bytes = resp_type.size if resp_type else 1
        resp = await self.send_command(
            message=request.encode(),
            n_resp_bytes=n_resp_bytes,
            timeout=3
        )
        if resp_type:
            size_resp = len(resp) if resp else 0
            if size_resp != resp_type.size:
                raise ValueError(
                    f'The size of the frame should be {resp_type.size}'
                    f' but is {size_resp}'
                )
        self._handle_data(resp)

    async def send_command(
        self,
        message: bytes,
        n_resp_bytes=0,
        timeout=3
    ) -> Optional[bytes]:
        """Send a command to the alarm and returns its response.

        Args:
            message (bytes):
                Message to send to the control panel

        Returns:
            bytes:
                Response of the message sent to the control panel or None
                otherwise
        """
        resp = None
        available = False

        if self.connected:
            async with self._lock:
                try:
                    resp = await self._send(message, n_resp_bytes, timeout)
                except asyncio.exceptions.TimeoutError:
                    _LOGGER.warning('Message not received on time')
                    raise
                except asyncio.IncompleteReadError:
                    _LOGGER.warning('Message not received.')
                    raise
                except EOFError:
                    _LOGGER.warning('Connection EOF')
                    raise
                except (OSError):
                    _LOGGER.warning('Connection failed')
                    await self._close_connection()
                    raise
                except BaseException:
                    _LOGGER.exception('Unexpected Error:')
                    raise
                else:
                    available = True
                    return resp
                finally:
                    await self._update_availability(available)
        else:
            await self._update_availability(available)
            raise ConnectionError('Not Connected')

    async def _update_availability(self, available: bool = True):
        if self.control_panel.availability.available != available:
            self.control_panel.availability.available = available
            for listener in self._control_panel_listeners:
                asyncio.create_task(
                    listener(0, self.control_panel.availability)
                )

    async def _connect(self):
        async with self._lock:
            _LOGGER.info('Connecting to control panel...')
            try:
                await self._close_connection()
                await self._open_connection()
            except asyncio.TimeoutError:
                _LOGGER.error('Connection to control panel timed out')
                raise
            except (OSError):
                _LOGGER.error('Connection to control panel failed')
                raise
            except BaseException:
                _LOGGER.error(
                    'Connection to control panel failed with unknown error')
                raise

    async def _close_connection(self):
        """Close the stream connection to the alarm."""
        if self._reader:
            if not self._reader.at_eof():
                self._reader.feed_eof()
        if self._writer:
            self._writer.close()
            await self._writer.wait_closed()

        self._reader = None
        self._writer = None

        self._connected = False

    async def _open_connection(self):
        """Open the stream connection to the alarm."""
        self._reader, self._writer = await asyncio.open_connection(
            self._ip,
            self._port
        )
        self._connected = True

    async def _send(
        self,
        message: bytes,
        n_resp_bytes: int = 0,
        timeout=3
    ) -> bytes:
        """Send a binary stream to the control panel and waits for its response.

        Args:
            message (bytes): Message to send to the control panel

        Returns:
            bytes: Response of the message sent to the control panel
        """
        if self.connected and self._writer and self._reader:
            # Ensure a clean buffer
            self._reader._buffer.clear()  # type: ignore

            # Send the command
            self._writer.write(message)
            await self._writer.drain()

            # Wait for a response
            if not self._reader.at_eof():
                if n_resp_bytes:
                    data = await asyncio.wait_for(
                        self._reader.readexactly(n_resp_bytes),
                        timeout=timeout
                    )
                else:
                    data = await asyncio.wait_for(
                        self._reader.read(32), timeout=timeout
                    )
                if not data:
                    raise asyncio.IncompleteReadError(data, 32)
                return data
            else:
                raise EOFError()
        else:
            raise ConnectionError('Not Connected')

    def _encode_key(self, key: str) -> int:
        # Is a number between 0 and 9
        if key.isdigit() and int(key) in range(0, 10):
            return int(key)
        elif key == '*':
            return 0x1B
        elif key == '#':
            return 0x1A
        else:
            _LOGGER.error('Unrecognized key %s', key)
            raise ValueError('Unrecognized key %s', key)

    def _encode_keys(self, keys: Union[str, List[str]]) -> bytes:
        return bytes(self._encode_key(key) for key in list(keys))

    def _validate_data(self, data: Optional[bytes]):
        if data:
            cs = checksum(data[0:-1])
            if cs != data[-1]:
                raise ValueError(
                    f'Data with invalid checksum. Received: {hex(data[-1])}, '
                    f'Calculated: {hex(cs)} ({data.hex()})'
                )
        else:
            raise RuntimeError('Checksum cannot be calculated without data')

    def _handle_data(self, data: Optional[bytes]):
        if data:

            self._validate_data(data)

            _LOGGER.debug('New Data: %s', to_hex(data))

            if self._is_status_msg(data):
                self._handle_status_msg(data)

            for listener in self._data_listeners:
                asyncio.create_task(listener(data))
        else:
            _LOGGER.warning('No Data Received!!!')

    @classmethod
    def _is_status_msg(cls, data: bytes):

        if bytes([data[0]]) != bytes.fromhex(StatusResponse.code):
            return False

        if len(data) != StatusResponse.size:
            raise ValueError(
                f'The size of the frame should be {StatusResponse.size}'
                f' but is {len(data)}'
            )

        return True

    def _handle_status_msg(self, data: bytes):
        resp = StatusResponse.decode(data, self._model)

        self._update_siren_status(resp.siren)
        self._update_output_status(resp.outs)
        self._update_area_status(resp.areas)
        self._update_zone_status(resp.zones)
        self.update_zone_enabled(resp.zones_en)
        self._update_time(resp.time)

    def _update_zone_status(self, zones: List[bool]):
        for i in range(self._model.n_zones):
            zone_number: Id = i + 1
            zone = self._control_panel.zones[zone_number]

            if zone.triggered != zones[i]:
                zone.triggered = zones[i]

                for listener in self._control_panel_listeners:
                    asyncio.create_task(listener(zone_number, zone))

                _LOGGER.info(
                    'Status of Zone %d changed to %d',
                    zone_number,
                    zone.triggered
                )

    def update_zone_enabled(self, zones_en: List[bool]):
        """Update zone enabled."""
        for i in range(self._model.n_zones):
            zone_number = i + 1
            zone = self._control_panel.zones[zone_number]

            if zone.enabled != zones_en[i]:
                zone.enabled = zones_en[i]

                for listener in self._control_panel_listeners:
                    asyncio.create_task(listener(zone_number, zone))

                _LOGGER.info(
                    'Zone enabling of Zone %d changed to %d',
                    zone_number,
                    zone.enabled
                )

    def _update_siren_status(self, status: bool):
        if self._control_panel.siren.on != status:
            self._control_panel.siren.on = status

            for listener in self._control_panel_listeners:
                asyncio.create_task(listener(0, self._control_panel.siren))

            _LOGGER.info('Siren changed to %d', self._control_panel.siren.on)

    def _update_output_status(self, outs: List[bool]):
        for i in range(self._model.n_outputs):
            out_number: Id = i + 1
            out = self._control_panel.outputs[out_number]

            if out.on != outs[i]:
                out.on = outs[i]

                for listener in self._control_panel_listeners:
                    asyncio.create_task(listener(out_number, out))

                _LOGGER.info('The output %d changed to %d', out_number, out.on)

    def _update_area_status(self, areas: List[ArmingMode]):

        for i in range(self._model.n_areas):
            area_number: Id = i + 1
            area = self._control_panel.areas[area_number]
            status_changed = False

            if areas[i] is ArmingMode.DISARMED:
                if area.mode is not ArmingMode.DISARMED:
                    area.mode = ArmingMode.DISARMED
                    status_changed = True
            elif areas[i] is ArmingMode.ARMED_AWAY:
                if area.mode is not ArmingMode.ARMED_AWAY:
                    status_changed = True
                    area.mode = ArmingMode.ARMED_AWAY
            elif areas[i] is not ArmingMode.ARMED_STAY:
                if area.mode is not ArmingMode.ARMED_STAY:
                    status_changed = True
                    area.mode = ArmingMode.ARMED_STAY

            if status_changed:
                for listener in self._control_panel_listeners:
                    asyncio.create_task(listener(area_number, area))

                _LOGGER.info(
                    'Status of Area %d changed to %s', area_number, area.mode
                )

    def _update_time(self, time: datetime.time):

        if self._control_panel.time.time != time:
            self._control_panel.time.time = time

            for listener in self._control_panel_listeners:
                asyncio.create_task(listener(0, self._control_panel.time))

            _LOGGER.info('Time updated %s', self._control_panel.time)

    async def _start_connection_task(self):
        self._conn_task = asyncio.create_task(self._connection_task())

    async def _stop_connection_task(self):
        if self._conn_task:
            self._conn_task.cancel()
            self._conn_task = None
        async with self._lock:
            await self._close_connection()

    async def _connection_task(self):
        while True:
            try:
                if not self.connected:
                    _LOGGER.info('Connecting')
                    await self._connect()
            except CancelledError:
                break
            except BaseException:
                _LOGGER.error('Connection failed')
            finally:
                await asyncio.sleep(3)



"""Control panel command line interface."""
import asyncio
import logging

from bosch.control_panel.cc880p.cli.cmds import handle_command
from bosch.control_panel.cc880p.cli.parser import Args
from bosch.control_panel.cc880p.cli.parser import get_args
from bosch.control_panel.cc880p.cp import CP
from bosch.control_panel.cc880p.models.cp import Area
from bosch.control_panel.cc880p.models.cp import Availability
from bosch.control_panel.cc880p.models.cp import ControlPanelEntity
from bosch.control_panel.cc880p.models.cp import CpVersion
from bosch.control_panel.cc880p.models.cp import Id
from bosch.control_panel.cc880p.models.cp import Output
from bosch.control_panel.cc880p.models.cp import Siren
from bosch.control_panel.cc880p.models.cp import Time
from bosch.control_panel.cc880p.models.cp import Zone
from bosch.control_panel.cc880p.utils import to_hex

logging.basicConfig(level=logging.WARNING)

prev_data = None


async def data_listener(data: bytes) -> bool:
    """Listen of any control panel change."""
    global prev_data
    if prev_data and prev_data[:-2] != data[:-2]:
        print('\nDifference:')
        print(f'\tBefore:\t{to_hex(prev_data)}')
        print(f'\tAfter:\t{to_hex(data)}')
    else:
        print('\nNo Changes:')
        print(to_hex(data))
    prev_data = data
    return True


async def cp_listener(id: Id, cp: ControlPanelEntity) -> bool:
    """Control panel listener."""
    if isinstance(cp, Zone):
        print(f'Zone {id} updated: {cp}')
    elif isinstance(cp, Output):
        print(f'Output {id} updated: {cp}')
    elif isinstance(cp, Siren):
        print(f'Siren updated: {cp}')
    elif isinstance(cp, Area):
        print(f'Area {id} updated: {cp}')
    elif isinstance(cp, Availability):
        print(f'Control Panel availability is: {cp}')
    elif isinstance(cp, Time):
        print(f'Control Panel time is: {cp}')

    return True


async def _wait_for_connection(cp: CP):
    while not cp.connected:
        await asyncio.sleep(1.0)


async def run_listen_mode(cp: CP):
    """Run the control panel in listen mode."""
    cp.add_data_listener(data_listener)
    cp.add_control_panel_listener(cp_listener)
    while True:
        try:
            await asyncio.wait_for(_wait_for_connection(cp), timeout=3.0)
            await cp.get_status()
        except BaseException:
            logging.exception('Error:')
        finally:
            await asyncio.sleep(1)


async def run_cmd_mode(cp: CP, args):
    """Run mode command."""
    await handle_command(cp, args)


async def run(loop):
    """Run the control panel command line."""
    args: Args = get_args()

    cp = CP(
        ip=args.connect,
        port=args.port,
        model=CpVersion.S16_V14.model(),
        installer_code=args.code,
        loop=loop,
    )

    await cp.start()

    if args.cmd:
        await run_cmd_mode(cp, args)
    else:
        await run_listen_mode(cp)

    await cp.stop()


def main():
    """Run main."""
    loop = asyncio.new_event_loop()
    loop.run_until_complete(run(loop))


if __name__ == '__main__':
    main()



"""Callbacks."""
from typing import Callable

from bosch.control_panel.cc880p.models.constants import Id
from bosch.control_panel.cc880p.models.cp import ControlPanelEntity


"""Control Panel Listener."""
ControlPanelListener = Callable[[Id, ControlPanelEntity], bool]
"""Data Listener."""
DataListener = Callable[[bytes], bool]
